package com.taka_2.ftp;

import java.io.*;
import java.net.*;
import java.util.*;

import com.taka_2.ftp.command.*;

/**
 *	FTPNCAgNX
 */
public class FTPClient
{
	// FTPT[oڑp|[gԍ
	private static final int FTP_PORT = 21;

	// ʐMp\Pbg
	private Session session;

	/**
	 *	T[oA[UApX[hw肵č\z
	 */
	public FTPClient(String serverName, String username, String password) throws FTPException, IOException
	{
		// T[oɐڑAŏ̉擾B
		session = new Session(serverName, FTP_PORT);
		new CommandReady().execCommand(session, "");

		// T[oɃOCB
		login(username, password);
	}

	/**
		[U[F؂s
		@param	user_name	[U[
		@param	password	pX[h
	*/
	private void login(String username, String password) throws FTPException, IOException
	{
		new CommandUser().execCommand(session, "USER " + username);
		new CommandPass().execCommand(session, "PASS " + password);
	}

	/**
		M^CvI
		@param	type	TransferType.TYPE_ASCII, TransferType.TYPE_BINARŶꂩ
	*/
	public void selectType(int type) throws FTPException, IOException
	{
		if(type == TransferType.TYPE_ASCII)
		{
			new CommandType().execCommand(session, "TYPE A");
		}
		else if(type ==TransferType.TYPE_BINARY)
		{
			new CommandType().execCommand(session, "TYPE I");
		}
		else
		{
			throw new FTPException("Illegal FTP Transfer Type");
		}
	}

	/**
	 *	T[ot@C̍폜
	 *
	 *	@param	fileName	폜t@C
	 */
	public void delete(String fileName) throws FTPException, IOException
	{
		new CommandDelete().execCommand(session, "DELE " + fileName);
	}

	/**
	 *	T[ot@C̃l[
	 *
	 *	@param	fileNameFrom	l[t@C
	 *	@param	fileNameTo		Vt@C
	 */
	public void rename(String fileNameFrom, String fileNameTo) throws FTPException, IOException
	{
		new CommandRenameFrom().execCommand(session, "RNFR " + fileNameFrom);
		new CommandRenameTo().execCommand(session, "RNTO " + fileNameTo);
	}

	/**
	 *	T[ot@C̃p[~bVύX
	 *
	 *	@param	fileName	p[~bVύXt@C
	 *	@param	permission	ύX̃p[~bV
	 */
	public void changePermission(String fileName, String permission) throws FTPException, IOException
	{
		new CommandChmod().execCommand(session, "SITE CHMOD " + permission + " " + fileName);
	}

	/**
	 *	JgfBNg̎擾
	 *
	 *	@return	directoryName	fBNg
	 */
	public String getCurrentDirectory() throws FTPException, IOException
	{
		String serverResponse = new CommandPwd().execCommand(session, "PWD");

		// fBNǵA_uNH[gɈ͂܂Ă
		int beginIndex = serverResponse.indexOf("\"");
		int endIndex  = serverResponse.indexOf("\"", beginIndex + 1);
		String result = serverResponse.substring(beginIndex + 1, endIndex);

		return(result);
	}

	/**
	 *	fBNgύX
	 *
	 *	@param	directoryName	ړfBNg
	 */
	public void changeDirectory(String directoryName) throws FTPException, IOException
	{
		new CommandChdir().execCommand(session, "CWD " + directoryName);
	}

	/**
	 *	fBNg쐬
	 *
	 *	@param	directoryName	쐬fBNg
	 */
	public void makeDirectory(String directoryName) throws FTPException, IOException
	{
		new CommandMkdir().execCommand(session, "MKD " + directoryName);
	}

	/**
	 *	fBNg폜
	 *
	 *	@param	directoryName	폜fBNg
	 */
	public void removeDirectory(String directoryName) throws FTPException, IOException
	{
		new CommandRmdir().execCommand(session, "RMD " + directoryName);
	}

	/**
	 *	NOOPR}h̔s
	 */
	public void noop() throws FTPException, IOException
	{
		new CommandNoop().execCommand(session, "NOOP");
	}

	/**
	 *	FTPڑ̏I
	 */
	public void quit() throws FTPException, IOException
	{
		new CommandQuit().execCommand(session, "QUIT");

		// ڑؒf
		session.close();
	}

	/**
	 *	HELPR}h̔s
	 */
	public String help() throws FTPException, IOException
	{
		return new CommandHelp().execCommand(session, "HELP");
	}

	/**
	 *	SYSTR}h̔s
	 */
	public String system() throws FTPException, IOException
	{
		return new CommandSystem().execCommand(session, "SYST");
	}

	/**
		wR}hAT[oɑΉĂ邩ǂ𒲂ׂ
		@param	command	ΏۃR}h
		@return	T[oΉĂtrueAłȂfalseԂ
	*/
	public boolean isCommandAvailable(String command_name) throws FTPException, IOException
	{
		return help().indexOf(command_name) != -1;
	}

	/**
		f[^MpNCAg|[g̃I[vƃT[oւ̒ʒm
	*/
	private ServerSocket openClientPort() throws FTPException, IOException
	{
		// f[^Mp\Pbg󂢂Ă|[gőҋ@
		ServerSocket ssocket = new ServerSocket(0);

		// [JIPAhX擾
		int[] local_ip = getLocalIPAddress();

		// T[oւ̃|[g̒ʒm
		int port = ssocket.getLocalPort();
		int high_port_number = port / 256;
		int low_port_number  = port % 256;
		new CommandPort().execCommand(session, "PORT " + local_ip[0] + "," + local_ip[1] + "," + local_ip[2] + "," + local_ip[3] +
							 "," + high_port_number + "," + low_port_number);

		return ssocket;
	}

	/**
	 *	[JzXgIPAhX擾
	 */
	private int[] getLocalIPAddress() throws UnknownHostException
	{
		int[] result = new int[4];

		String str_local_ip = InetAddress.getLocalHost().getHostAddress();
		StringTokenizer st = new StringTokenizer(str_local_ip, ".");

		for(int i=0; st.hasMoreTokens(); i++)
			result[i] = Integer.parseInt(st.nextToken());

		return result;
	}

	/**
		t@CXg̎擾
		@return	t@CXg(ls -alɂ)
	*/
	public String[] getFileList() throws FTPException, IOException
	{
		// t@CXg
		Vector result_array = new Vector();

		// NCAg̏
		selectType(TransferType.TYPE_ASCII);
		ServerSocket ssocket = openClientPort();

		// XgR}h̑M
		new CommandList().execCommand(session, "NLST -alL");

		try
		{
			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// Xg[̍쐬
			BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));

			// Xg̎擾
			String buf;
			while((buf = reader.readLine()) != null)
			{
				result_array.add(buf);
			}

			// Xg[
			reader.close();
		}
		catch(IOException io_ex)
		{
			throw new FTPException("Failed to get File List.");
		}

		// Transfer Complete̎M
		new CommandTransfer().execCommand(session, "");

		// Vector  zϊ
		String[] result = new String[result_array.size()];
		result_array.toArray(result);

		return result;
	}

	/**
		AXL[t@C̃Abv[h
		@param	fileName	Mt@C
		@param	kanjiEncoding	{R[h(Shift_JIS, EUC-JP, ISO-2022-JP...)
	*/
	public void putAsciiFile(String fileName, String kanjiEncoding) throws FTPException, IOException
	{
		// AXL[[hɃZbg
		selectType(TransferType.TYPE_ASCII);

		// NCAg̏
		ServerSocket ssocket = openClientPort();

		// t@CMR}h̔s
		new CommandStore().execCommand(session, "STOR " + fileName);

		try
		{
			// Mt@C̎擾
			StringBuffer src = new StringBuffer();
			BufferedReader reader = new BufferedReader(new FileReader(fileName));

			String buf;
			while((buf = reader.readLine()) != null)
			{
				src.append(buf + System.getProperty("line.separator"));
			}

			// Xg[
			reader.close();

			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// AXL[t@C̑M
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), kanjiEncoding));
			writer.write(src.toString(), 0, src.length());

			// Xg[
			writer.close();
		}
		catch(IOException io_ex)
		{
			throw new FTPException("Can't send " + fileName);
		}

		// Transfer Complete̎M
		new CommandTransfer().execCommand(session, "");
	}

	/**
		oCit@C̃Abv[h
		@param	file_name	Mt@C
	*/
	public void putBinaryFile(String fileName) throws FTPException, IOException
	{
		// oCi[hɃZbg
		selectType(TransferType.TYPE_BINARY);

		// NCAg̏
		ServerSocket ssocket = openClientPort();

		// t@CMR}h̔s
		new CommandStore().execCommand(session, "STOR " + fileName);

		try
		{
			// Mt@C̎擾
			File src_file = new File(fileName);
			BufferedInputStream input = new BufferedInputStream(new FileInputStream(src_file));

			byte src[] = new byte[(int)src_file.length()];
			input.read(src);

			// Xg[
			input.close();

			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// oCit@C̑M
			BufferedOutputStream output = new BufferedOutputStream(socket.getOutputStream());
			output.write(src, 0, src.length);

			// Xg[
			output.close();
		}
		catch(IOException io_ex)
		{
			throw new FTPException("Can't send " + fileName);
		}

		// Transfer Complete̎M
		new CommandTransfer().execCommand(session, "");
	}

	/**
		t@C擾̏(M) - NCAg|[gJĂ邱ƂK{
		@param	file_name	擾Ώۃt@C
		@return	Ώۃt@C̃t@CTCY
	*/
	private int prepareGetFile(String fileName) throws FTPException, IOException
	{
		// t@C擾R}h̔s
		String serverResponse = new CommandRetr().execCommand(session, "RETR " + fileName);

		// T[ỏAt@CTCYoĕԋp
		return getFileSizeFromMessage(serverResponse);
	}

	/**
	 *	T[obZ[Wt@CTCY擾(M)
	 *
	 *	@param	serverResponse	T[õbZ[W
	 *	@return	bZ[W蒊oꂽt@CTCY
	 */
	private int getFileSizeFromMessage(String serverResponse)
	{
		int result = 0;

		int beginIndex = serverResponse.lastIndexOf('(') + 1;
		int endIndex   = serverResponse.lastIndexOf("bytes") - 1;
		result = Integer.parseInt(serverResponse.substring(beginIndex, endIndex));

		return result;
	}

	/**
		wt@C̃t@CTCY擾
		@param	file_name	t@C
		@return	t@CTCY
	*/
	public int getFileSize(String fileName) throws FTPException, IOException
	{
		// NCAg̏
		ServerSocket ssocket = openClientPort();

		// RETR𑗐MāAt@CTCY擾B
		int result = prepareGetFile(fileName);

		// MȂŁATransfer Complete̎Ms
		ssocket.close();
		new CommandTransfer().execCommand(session, "");

		return result;
	}

	/**
		AXL[t@C(eLXgt@C)̎擾
		@param	file_name	擾t@C
		@param	kanji_encoding	{R[h(Shift_JIS, EUC-JP, ISO-2022-JP...)
		@return	t@C̓e
	*/
	public String getAsciiFile(String fileName, String kanjiEncoding) throws FTPException, IOException
	{
		// ʕێpϐ
		StringBuffer buffer = new StringBuffer();

		// AXL[[hɃZbg
		selectType(TransferType.TYPE_ASCII);

		// t@C擾̏ (|[gI[v & RETR)
		ServerSocket ssocket = openClientPort();
		prepareGetFile(fileName);

		try
		{
			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();

			// AXL[f[^̎擾
			BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream(), kanjiEncoding));
			String buf;
			while((buf = reader.readLine()) != null)
			{
				buffer.append(buf + System.getProperty("line.separator"));
			}
			reader.close();
		}
		catch(IOException io_ex)
		{
			new FTPException("Can't get " + fileName);
		}

		// Transfer Complete̎M
		new CommandTransfer().execCommand(session, "");

		return(buffer.toString());
	}

	/**
		oCit@C̎擾
		@param	buffer		f[^󂯎obt@
		@param	file_name	擾t@C
		@param	length		擾f[^TCY
		@return	Mf[^TCY
	*/
	public int getBinaryFile(String fileName, byte[] buffer, int length) throws FTPException, IOException
	{
		// oCi[hɃZbg
		selectType(TransferType.TYPE_BINARY);

		// t@C擾̏ (|[gI[v & RETR)
		ServerSocket ssocket = openClientPort();
		int fileSize = prepareGetFile(fileName);

		// ǂݍ񂾃oCg
		int readLength = 0;

		try
		{
			// T[o̐ڑ̎󂯕t
			Socket socket = ssocket.accept();
			BufferedInputStream input = new BufferedInputStream(socket.getInputStream());

			// oCif[^̎擾
			while(readLength < length)
			{
				readLength += input.read(buffer, readLength, (length - readLength));
			}
			input.close();
		}
		catch(IOException io_ex)
		{
			throw new FTPException("Can't get " + fileName);
		}

		// Transfer Complete̎M
		new CommandTransfer().execCommand(session, "");

		return readLength;
	}
}